@@ -1,4 +1,4 @@
-/* The GIMP -- an image manipulation program
+/* GIMP - The GNU Image Manipulation Program
  * Copyright (C) 1995 Spencer Kimball and Peter Mattis
  *
  * This program is free software; you can redistribute it and/or modify
@@ -20,17 +20,25 @@
 
 #include <gtk/gtk.h>
 
+#include "libgimpbase/gimpbase.h"
+#include "libgimpmath/gimpmath.h"
+
 #include "display-types.h"
 
-#include "core/gimp-utils.h"
+#include "base/tile-manager.h"
+
+#include "core/gimpcontext.h"
 #include "core/gimpgrid.h"
+#include "core/gimpguide.h"
 #include "core/gimpimage.h"
-#include "core/gimpimage-guides.h"
 #include "core/gimplist.h"
+#include "core/gimpprojection.h"
+#include "core/gimpsamplepoint.h"
 
 #include "vectors/gimpstroke.h"
 #include "vectors/gimpvectors.h"
 
+#include "widgets/gimprender.h"
 #include "widgets/gimpwidgets-utils.h"
 
 #include "gimpcanvas.h"
@@ -46,6 +54,9 @@
 
 static GdkGC * gimp_display_shell_get_grid_gc (GimpDisplayShell *shell,
                                                GimpGrid         *grid);
+static GdkGC * gimp_display_shell_get_pen_gc  (GimpDisplayShell *shell,
+                                               GimpContext      *context,
+                                               GimpActiveColor   active);
 
 
 /*  public functions  */
@@ -55,41 +66,40 @@ gimp_display_shell_draw_guide (GimpDisplayShell *shell,
                                GimpGuide        *guide,
                                gboolean          active)
 {
-  gint  x1, x2;
-  gint  y1, y2;
-  gint  x, y;
-  gint  w, h;
+  gint  position;
+  gint  x1, x2, y1, y2;
+  gint  x, y, w, h;
 
   g_return_if_fail (GIMP_IS_DISPLAY_SHELL (shell));
-  g_return_if_fail (guide != NULL);
+  g_return_if_fail (GIMP_IS_GUIDE (guide));
 
-  if (guide->position < 0)
+  position = gimp_guide_get_position (guide);
+
+  if (position < 0)
     return;
 
   gimp_display_shell_transform_xy (shell, 0, 0, &x1, &y1, FALSE);
   gimp_display_shell_transform_xy (shell,
-                                   shell->gdisp->gimage->width,
-                                   shell->gdisp->gimage->height,
+                                   shell->display->image->width,
+                                   shell->display->image->height,
                                    &x2, &y2, FALSE);
 
   gdk_drawable_get_size (shell->canvas->window, &w, &h);
 
-  if (x1 < 0) x1 = 0;
-  if (y1 < 0) y1 = 0;
-  if (x2 > w) x2 = w;
-  if (y2 > h) y2 = h;
+  x1 = MAX (x1, 0);
+  y1 = MAX (y1, 0);
+  x2 = MIN (x2, w);
+  y2 = MIN (y2, h);
 
-  switch (guide->orientation)
+  switch (gimp_guide_get_orientation (guide))
     {
     case GIMP_ORIENTATION_HORIZONTAL:
-      gimp_display_shell_transform_xy (shell,
-                                       0, guide->position, &x, &y, FALSE);
+      gimp_display_shell_transform_xy (shell, 0, position, &x, &y, FALSE);
       y1 = y2 = y;
       break;
 
     case GIMP_ORIENTATION_VERTICAL:
-      gimp_display_shell_transform_xy (shell,
-                                       guide->position, 0, &x, &y, FALSE);
+      gimp_display_shell_transform_xy (shell, position, 0, &x, &y, FALSE);
       x1 = x2 = x;
       break;
 
@@ -112,14 +122,14 @@ gimp_display_shell_draw_guides (GimpDisplayShell *shell)
     {
       GList *list;
 
-      for (list = shell->gdisp->gimage->guides;
+      for (list = shell->display->image->guides;
            list;
            list = g_list_next (list))
-	{
-	  gimp_display_shell_draw_guide (shell,
+        {
+          gimp_display_shell_draw_guide (shell,
                                          (GimpGuide *) list->data,
                                          FALSE);
-	}
+        }
     }
 }
 
@@ -143,7 +153,7 @@ gimp_display_shell_draw_grid (GimpDisplayShell   *shell,
 
 #define CROSSHAIR 2
 
-      grid = GIMP_GRID (shell->gdisp->gimage->grid);
+      grid = GIMP_GRID (shell->display->image->grid);
       if (! grid)
         return;
 
@@ -154,8 +164,8 @@ gimp_display_shell_draw_grid (GimpDisplayShell   *shell,
       x2 = area->x + area->width;
       y2 = area->y + area->height;
 
-      width  = shell->gdisp->gimage->width;
-      height = shell->gdisp->gimage->height;
+      width  = shell->display->image->width;
+      height = shell->display->image->height;
 
       x_offset = grid->xoffset;
       while (x_offset > 0)
@@ -291,6 +301,135 @@ gimp_display_shell_draw_grid (GimpDisplayShell   *shell,
     }
 }
 
+void
+gimp_display_shell_draw_pen (GimpDisplayShell  *shell,
+                             const GimpVector2 *points,
+                             gint               num_points,
+                             GimpContext       *context,
+                             GimpActiveColor    color,
+                             gint               width)
+{
+  GimpCanvas  *canvas;
+  GdkGC       *gc;
+  GdkGCValues  values;
+  GdkPoint    *coords;
+
+  g_return_if_fail (GIMP_IS_DISPLAY_SHELL (shell));
+  g_return_if_fail (GIMP_IS_CONTEXT (context));
+  g_return_if_fail (num_points == 0 || points != NULL);
+
+  canvas = GIMP_CANVAS (shell->canvas);
+
+  coords = g_new (GdkPoint, MAX (2, num_points));
+
+  gimp_display_shell_transform_points (shell,
+                                       (const gdouble *) points, coords,
+                                       num_points, FALSE);
+
+  if (num_points == 1)
+    {
+      coords[1] = coords[0];
+      num_points = 2;
+    }
+
+  gc = gimp_display_shell_get_pen_gc (shell, context, color);
+
+  values.line_width = MAX (1, width);
+  gdk_gc_set_values (gc, &values, GDK_GC_LINE_WIDTH);
+
+  gimp_canvas_set_custom_gc (canvas, gc);
+
+  gimp_canvas_draw_lines (canvas, GIMP_CANVAS_STYLE_CUSTOM, coords, num_points);
+
+  gimp_canvas_set_custom_gc (canvas, NULL);
+
+  g_free (coords);
+}
+
+void
+gimp_display_shell_draw_sample_point (GimpDisplayShell *shell,
+                                      GimpSamplePoint  *sample_point,
+                                      gboolean          active)
+{
+  GimpCanvasStyle style;
+  gdouble         x, y;
+  gint            x1, x2;
+  gint            y1, y2;
+  gint            w, h;
+
+  g_return_if_fail (GIMP_IS_DISPLAY_SHELL (shell));
+  g_return_if_fail (sample_point != NULL);
+
+  if (sample_point->x < 0)
+    return;
+
+  gimp_display_shell_transform_xy_f (shell,
+                                     sample_point->x + 0.5,
+                                     sample_point->y + 0.5,
+                                     &x, &y, FALSE);
+
+  x1 = floor (x - GIMP_SAMPLE_POINT_DRAW_SIZE);
+  x2 = ceil  (x + GIMP_SAMPLE_POINT_DRAW_SIZE);
+  y1 = floor (y - GIMP_SAMPLE_POINT_DRAW_SIZE);
+  y2 = ceil  (y + GIMP_SAMPLE_POINT_DRAW_SIZE);
+
+  gdk_drawable_get_size (shell->canvas->window, &w, &h);
+
+  if (x < - GIMP_SAMPLE_POINT_DRAW_SIZE   ||
+      y < - GIMP_SAMPLE_POINT_DRAW_SIZE   ||
+      x > w + GIMP_SAMPLE_POINT_DRAW_SIZE ||
+      y > h + GIMP_SAMPLE_POINT_DRAW_SIZE)
+    return;
+
+  if (active)
+    style = GIMP_CANVAS_STYLE_SAMPLE_POINT_ACTIVE;
+  else
+    style = GIMP_CANVAS_STYLE_SAMPLE_POINT_NORMAL;
+
+#define HALF_SIZE (GIMP_SAMPLE_POINT_DRAW_SIZE / 2)
+
+  gimp_canvas_draw_line (GIMP_CANVAS (shell->canvas), style,
+                         x, y1, x, y1 + HALF_SIZE);
+  gimp_canvas_draw_line (GIMP_CANVAS (shell->canvas), style,
+                         x, y2 - HALF_SIZE, x, y2);
+
+  gimp_canvas_draw_line (GIMP_CANVAS (shell->canvas), style,
+                         x1, y, x1 + HALF_SIZE, y);
+  gimp_canvas_draw_line (GIMP_CANVAS (shell->canvas), style,
+                         x2 - HALF_SIZE, y, x2, y);
+
+  gimp_canvas_draw_arc (GIMP_CANVAS (shell->canvas), style,
+                        FALSE,
+                        x - HALF_SIZE, y - HALF_SIZE,
+                        GIMP_SAMPLE_POINT_DRAW_SIZE,
+                        GIMP_SAMPLE_POINT_DRAW_SIZE,
+                        0, 64 * 270);
+
+  gimp_canvas_draw_text (GIMP_CANVAS (shell->canvas), style,
+                         x + 2, y + 2,
+                         "%d",
+                         g_list_index (shell->display->image->sample_points,
+                                       sample_point) + 1);
+}
+
+void
+gimp_display_shell_draw_sample_points (GimpDisplayShell *shell)
+{
+  g_return_if_fail (GIMP_IS_DISPLAY_SHELL (shell));
+
+  if (gimp_display_shell_get_show_sample_points (shell))
+    {
+      GList *list;
+
+      for (list = shell->display->image->sample_points;
+           list;
+           list = g_list_next (list))
+        {
+          gimp_display_shell_draw_sample_point (shell, list->data, FALSE);
+        }
+    }
+}
+
 void
 gimp_display_shell_draw_vector (GimpDisplayShell *shell,
                                 GimpVectors      *vectors)
@@ -307,26 +446,16 @@ gimp_display_shell_draw_vector (GimpDisplayShell *shell,
 
       coords = gimp_stroke_interpolate (stroke, 1.0, &closed);
 
-      if (coords && coords->len)
+      if (coords && coords->len > 0)
         {
-          GimpCoords *coord;
-          GdkPoint   *gdk_coords;
-          gint        i;
-          gint        sx, sy;
+          GdkPoint *gdk_coords = g_new (GdkPoint, coords->len);
 
-          gdk_coords = g_new (GdkPoint, coords->len);
-
-          for (i = 0; i < coords->len; i++)
-            {
-              coord = &g_array_index (coords, GimpCoords, i);
-
-              gimp_display_shell_transform_xy (shell,
-                                               coord->x, coord->y,
-                                               &sx, &sy,
+          gimp_display_shell_transform_coords (shell,
+                                               &g_array_index (coords,
+                                                               GimpCoords, 0),
+                                               gdk_coords,
+                                               coords->len,
                                                FALSE);
-              gdk_coords[i].x = sx;
-              gdk_coords[i].y = sy;
-            }
 
           gimp_canvas_draw_lines (GIMP_CANVAS (shell->canvas),
                                   GIMP_CANVAS_STYLE_XOR,
@@ -349,15 +478,15 @@ gimp_display_shell_draw_vectors (GimpDisplayShell *shell)
     {
       GList *list;
 
-      for (list = GIMP_LIST (shell->gdisp->gimage->vectors)->list;
+      for (list = GIMP_LIST (shell->display->image->vectors)->list;
            list;
            list = list->next)
-	{
+        {
           GimpVectors *vectors = list->data;
 
           if (gimp_item_get_visible (GIMP_ITEM (vectors)))
             gimp_display_shell_draw_vector (shell, vectors);
-	}
+        }
     }
 }
 
@@ -378,23 +507,39 @@ gimp_display_shell_draw_area (GimpDisplayShell *shell,
                               gint              w,
                               gint              h)
 {
-  gint sx, sy;
-  gint sw, sh;
+  GimpProjection *proj = shell->display->image->projection;
+  TileManager    *tiles;
+  gint            level;
+  gint            level_width;
+  gint            level_height;
+  gint            sx, sy;
+  gint            sw, sh;
 
-  g_return_if_fail (GIMP_IS_DISPLAY_SHELL (shell));
+  g_return_if_fail (GIMP_IS_DISPLAY_SHELL (shell) &&
+                    GIMP_IS_PROJECTION (proj));
+
+  level = gimp_projection_get_level (proj, shell->scale_x, shell->scale_y);
+
+  tiles = gimp_projection_get_tiles_at_level (proj, level);
+
+  level_width  = tile_manager_width (tiles);
+  level_height = tile_manager_height (tiles);
 
   /*  the image's size in display coordinates  */
-  sx = shell->disp_xoffset > 0 ? shell->disp_xoffset : - shell->offset_x;
-  sy = shell->disp_yoffset > 0 ? shell->disp_yoffset : - shell->offset_y;
-  sw = SCALEX (shell, shell->gdisp->gimage->width);
-  sh = SCALEY (shell, shell->gdisp->gimage->height);
+  sx = shell->disp_xoffset - shell->offset_x;
+  sy = shell->disp_yoffset - shell->offset_y;
+
+  /* SCALE[XY] with pyramid level taken into account. */
+  sw = PROJ_ROUND (level_width  * (shell->scale_x * (1 << level)));
+  sh = PROJ_ROUND (level_height * (shell->scale_y * (1 << level)));
 
   /*  check if the passed in area intersects with
    *  both the display and the image
    */
   if (gimp_rectangle_intersect (x, y, w, h,
                                 0, 0, shell->disp_width,  shell->disp_height,
-                                &x, &y, &w, &h) &&
+                                &x, &y, &w, &h)
+      &&
       gimp_rectangle_intersect (x, y, w, h,
                                 sx, sy, sw, sh,
                                 &x, &y, &w, &h))
@@ -408,38 +553,29 @@ gimp_display_shell_draw_area (GimpDisplayShell *shell,
 
       if (shell->highlight)
         {
-          rect.x      = SCALEX (shell, shell->highlight->x);
-          rect.y      = SCALEY (shell, shell->highlight->y);
-          rect.width  = SCALEX (shell, shell->highlight->width);
-          rect.height = SCALEY (shell, shell->highlight->height);
+          rect.x      = ceil  (shell->scale_x * shell->highlight->x);
+          rect.y      = ceil  (shell->scale_y * shell->highlight->y);
+          rect.width  = floor (shell->scale_x * shell->highlight->width);
+          rect.height = floor (shell->scale_y * shell->highlight->height);
         }
 
       /*  display the image in RENDER_BUF_WIDTH x RENDER_BUF_HEIGHT
        *  sized chunks
        */
-      for (i = y; i < y2; i += GIMP_DISPLAY_SHELL_RENDER_BUF_HEIGHT)
+      for (i = y; i < y2; i += GIMP_RENDER_BUF_HEIGHT)
         {
-          for (j = x; j < x2; j += GIMP_DISPLAY_SHELL_RENDER_BUF_WIDTH)
+          for (j = x; j < x2; j += GIMP_RENDER_BUF_WIDTH)
             {
               gint dx, dy;
 
-              dx = MIN (x2 - j, GIMP_DISPLAY_SHELL_RENDER_BUF_WIDTH);
-              dy = MIN (y2 - i, GIMP_DISPLAY_SHELL_RENDER_BUF_HEIGHT);
+              dx = MIN (x2 - j, GIMP_RENDER_BUF_WIDTH);
+              dy = MIN (y2 - i, GIMP_RENDER_BUF_HEIGHT);
 
               gimp_display_shell_render (shell,
                                          j - shell->disp_xoffset,
                                          i - shell->disp_yoffset,
                                          dx, dy,
                                          shell->highlight ? &rect : NULL);
-
-#ifdef STRESS_TEST
-              /* Invalidate the projection just after we render it! */
-              gimp_image_invalidate_without_render (shell->gdisp->gimage,
-                                                    j - shell->disp_xoffset,
-                                                    i - shell->disp_yoffset,
-                                                    dx, dy,
-                                                    0, 0, 0, 0);
-#endif
             }
         }
     }
@@ -482,10 +618,55 @@ gimp_display_shell_get_grid_gc (GimpDisplayShell *shell,
                                                      GDK_GC_JOIN_STYLE));
 
   gimp_rgb_get_gdk_color (&grid->fgcolor, &fg);
-  gimp_rgb_get_gdk_color (&grid->bgcolor, &bg);
-
   gdk_gc_set_rgb_fg_color (shell->grid_gc, &fg);
+
+  gimp_rgb_get_gdk_color (&grid->bgcolor, &bg);
   gdk_gc_set_rgb_bg_color (shell->grid_gc, &bg);
 
   return shell->grid_gc;
 }
+
+static GdkGC *
+gimp_display_shell_get_pen_gc (GimpDisplayShell *shell,
+                               GimpContext      *context,
+                               GimpActiveColor   active)
+{
+  GdkGCValues  values;
+  GimpRGB      rgb;
+  GdkColor     color;
+
+  if (shell->pen_gc)
+    return shell->pen_gc;
+
+  values.line_style = GDK_LINE_SOLID;
+  values.cap_style  = GDK_CAP_ROUND;
+  values.join_style = GDK_JOIN_ROUND;
+
+  shell->pen_gc = gdk_gc_new_with_values (shell->canvas->window,
+                                          &values, (GDK_GC_LINE_STYLE |
+                                                    GDK_GC_CAP_STYLE  |
+                                                    GDK_GC_JOIN_STYLE));
+
+  switch (active)
+    {
+    case GIMP_ACTIVE_COLOR_FOREGROUND:
+      gimp_context_get_foreground (context, &rgb);
+      break;
+
+    case GIMP_ACTIVE_COLOR_BACKGROUND:
+      gimp_context_get_background (context, &rgb);
+      break;
+    }
+
+  gimp_rgb_get_gdk_color (&rgb, &color);
+  gdk_gc_set_rgb_fg_color (shell->pen_gc, &color);
+
+  g_object_add_weak_pointer (G_OBJECT (shell->pen_gc),
+                             (gpointer) &shell->pen_gc);
+
+  g_signal_connect_object (context, "notify",
+                           G_CALLBACK (g_object_unref),
+                           shell->pen_gc, G_CONNECT_SWAPPED);
+
+  return shell->pen_gc;
+}
